/* Conway's game of life.

   Copyright 2006, Akkana Peck, akkana@shallowsky.com.
   This is free software: you may use it under the terms of the GPL.
 */

#include "grid.h"
#include "xgrid.h"

#include <stdio.h>
#include <stdlib.h>    // for exit()

#define INT_ARG(arg) \
   if (argc < 2) Usage(progname); \
   arg = atoi(argv[1]); \
   argc -= 2; \
   argv += 2

int InitDensity = 10;       /* initial density (percent) */
int InitGroupSize = 20;     /* size of initial area to be filled */

void Usage(const char* progname)
{
    fprintf(stderr,
            "Usage: %s [-V] [-c [-t msec]] [-d density] [-g initial_group_size]\n\t\t[-x x_cells] [-y y_cells]\n",
            progname);
    exit(1);
}

void ParseArgs(int argc, char **argv)
{
    const char* progname = *argv;
    --argc;
    ++argv;
    while (argc > 0 && argv[0][0] == '-')
    {
        switch (argv[0][1])
        {
          case 'V':
              Verbose = 1;
              --argc;
              ++argv;
              break;
          case 'c':
              Continuous = 1;
              --argc;
              ++argv;
              break;
          case 't':
              INT_ARG(TickTime);
              TickTime *= 1000;
              break;
          case 'd':
              INT_ARG(InitDensity);
              break;
          case 'g':
              INT_ARG(InitGroupSize);
              break;
          case 'x':
              INT_ARG(GridX);
              break;
          case 'y':
              INT_ARG(GridY);
              break;
          default:
              Usage(progname);
        }
    }
}

#define GROUP InitGroupSize

void PopulateGrid()
{
    int startx = random() % (GridX-GROUP*2) + GROUP;
    int starty = random() % (GridY-GROUP*2) + GROUP;
    int i, j;

    if (Verbose)
        printf("start @ (%d, %d)\n", startx-GROUP, starty-GROUP);
    for (i = startx-GROUP; i < startx+GROUP; ++i)
        for (j = starty-GROUP; j < starty+GROUP; ++j)
            if (random() % 100 < InitDensity)
                Grid[i][j] = 1;
}

main(int argc, char **argv)
{
    GridX = 150;
    GridY = 150;

    ParseArgs(argc, argv);

    InitWindow();

    PopulateGrid();

    /* Initialize colors */
    AllocColor("black");
    AllocColor("yellow");

    grid_main();
}

int Step()
{
    /* Count nearest neighbors */
    int i, j;
    int changed = 0;
    int countGrid[GridX][GridY];
    for (i=0; i<GridX; ++i)
        for (j=0; j<GridY; ++j)
        {
            int ii, jj;
            countGrid[i][j] = 0;
            for (ii=i-1; ii <= i+1; ++ii)
                for (jj=j-1; jj <= j+1; ++jj) {
                    int ti, tj;
                    if (ii == i && jj == j)
                        continue;
                    /* Periodic boundary conditions */
                    if (ii < 0) ti = GridX-1;
                    else if (ii >= GridX) ti = 0;
                    else ti = ii;
                    if (jj < 0) tj = GridY-1;
                    else if (jj >= GridY) tj = 0;
                    else tj = jj;

                    countGrid[i][j] += Grid[ti][tj];
                }
        }

    /* Finally, the Life rule */
    for (i=0; i<GridX; ++i)
        for (j=0; j<GridY; ++j) {
            int old = Grid[i][j];
            if (countGrid[i][j] == 2)
                Grid[i][j] = (Grid[i][j] == 1 ? 1 : 0);
            else if (countGrid[i][j] == 3)
                Grid[i][j] = 1;
            else
                Grid[i][j] = 0;
            if (Grid[i][j] != old)
                ++changed;
        }
    return changed;
}




