/* X11-specific CA routines */

#include "xgrid.h"
#include "grid.h"

#include <X11/keysym.h>
#include <X11/Xutil.h>

#include <stdio.h>
#include <stdlib.h>    /* for getenv */

/* Globals representing the grid */
unsigned char Grid[MAXGRID][MAXGRID];
int GridX = 0;
int GridY = 0;

int NumColors = 0;
unsigned long Colors[MAX_COLORS];

/* Statics used for drawing */
static int Scale = 4;           /* Size of a cell, in pixels */
static int XWinSize;            /* Size of the window (Scale*GridX) */
static int YWinSize;            /* Size of the window (Scale*GridY) */
static Display *display;
static int screen;
static Window win;
static GC gc = 0;

/* Loop continuously? */
int Continuous = 0;
/* How fast?  (usec) */
int TickTime = 0;

int Verbose = 0;

void InitWindow()
{
    XWinSize = GridX * Scale;
    YWinSize = GridY * Scale;
    srandom(getpid());
    if ((display = XOpenDisplay(getenv("DISPLAY"))) == 0)
    {
        fprintf(stderr, "Can't open display: %s\n", getenv("DISPLAY"));
        exit(1);
    }
    screen = DefaultScreen(display);

    win = XCreateSimpleWindow(display, RootWindow(display, screen),
                              0, 0, XWinSize, YWinSize, 3,
                              WhitePixel(display, screen),
                              BlackPixel(display, screen));
    if (!win)
    {
        fprintf(stderr, "Can't create window\n");
        exit(1);
    }

    XSelectInput(display, win, EventMask);

    XMapWindow(display, win);
}

void ChangeColor(unsigned long c)
{
    static unsigned long lastColor = 0;

    if (!gc)
    {
        gc = XCreateGC(display, win,  0, 0);
        if (!gc) {
            printf("ChangeColor: no GC!\n");
            return;
        }
        lastColor = c + 1;  // anything that's not *c
    }

    if (c != lastColor)
    {
        XGCValues values;
        values.foreground = c;
        XChangeGC(display, gc, GCForeground, &values);
        lastColor = c;
    }
}

void RedrawScreen()
{
    int i, j;
    int rightedge = GridX * Scale;
    int bottomedge = GridY * Scale;
    for (i=0; i < GridX; ++i)
        for (j=0; j < GridY; ++j)
        {
            int x = i * Scale;
            int y = j * Scale;
            if (Grid[i][j] < NumColors)
                ChangeColor(Colors[Grid[i][j]]);
            else {
                fprintf(stderr, "Bad value! (%d, %d): %d\n", i, j, Grid[i][j]);
                ChangeColor(WhitePixel(display, screen));
            }
            XFillRectangle(display, win, gc, x, y, Scale, Scale);
        }
#if 0
    XClearArea(display, win,
               rightedge+1, 0, XWinSize-rightedge-1, YWinSize, 0);
    XClearArea(display, win,
               0, bottomedge+1, XWinSize, YWinSize-bottomedge-1, 0);
    ChangeColor(WhitePixel(display, screen));
    XDrawLine(display, win, gc, rightedge, 0, rightedge, bottomedge);
    XDrawLine(display, win, gc, 0, bottomedge, rightedge, bottomedge);
#endif
}

int AllocColor(char* colorname)
{
    if (NumColors >= MAX_COLORS-1)
        return 0;

    XColor color, exact;
    if (XAllocNamedColor(display,
                          DefaultColormap(display, screen),
                          colorname, &color, &exact))
        return (Colors[NumColors++] = color.pixel);
    fprintf(stderr, "Can't allocate %s\n", colorname);
    return (Colors[NumColors++] = WhitePixel(display, screen));
}

/* Return how many cells changed in the last step. */
int Iterate(int steps)
{
    int changed;
    if (Verbose && steps > 1)
        printf("Iterating %d steps\n", steps);
    while (steps--)
        changed = Step();
    if (Verbose)
        printf("%d cells changed\n", changed);
    return changed;
}

void grid_main()
{
    while (1) {
        XEvent event;

        if (Continuous) {
            if (Iterate(1) == 0) {
                Continuous = 0;
                if (Verbose) printf("Reached a stable state\n");
            }
            RedrawScreen();
            usleep(TickTime);
            if ( XCheckWindowEvent(display, win, EventMask, &event) )
                if (HandleEvent(&event) != 0)
                    return;
        }
        else {
            XNextEvent(display, &event);
            if (HandleEvent(&event) != 0)
                return;
        }
    }
}

int HandleEvent(XEvent* event)
{
    char buffer[20];
    KeySym keysym;
    XComposeStatus compose;
    time_t sec;
    int i, j;
    static int lastX = -1;
    static int lastY = -1;
    extern double WIDESCALE, NARROWSCALE;

    switch (event->type)
    {
    case Expose:
        RedrawScreen();
        break;
    case ConfigureNotify:
        /* For now, keep scale the same and change the grid size */
        GridX = event->xconfigure.width / Scale;
        GridY = event->xconfigure.height / Scale;
        XWinSize = GridX * Scale;
        YWinSize = GridY * Scale;
        /* XWinSize and YWinSize get clipped to nearest multiple of Scale */
        if (Verbose)
            printf("ConfigureNotify: grid now (%d, %d)\n", GridX, GridY);
        XClearWindow(display, win);
        break;
    case ButtonPress:
        lastX = event->xbutton.x;
        lastY = event->xbutton.y;
        break;
    case ButtonRelease:
        i = event->xbutton.x * GridX / XWinSize;
        j = event->xbutton.y * GridY / YWinSize;
        Grid[i][j] = (Grid[i][j] + 1) % NumColors;
        RedrawScreen();
        break;
    case KeyPress:
        XLookupString(&(event->xkey), buffer, sizeof buffer,
                      &keysym, &compose);
        switch (keysym)
        {
        case XK_q:
            return -1;
        case XK_space:
            Continuous = 0;
            Iterate(1);
            RedrawScreen();
            break;
        case XK_r:
        case XK_s:
        case XK_Return:
            Iterate(10);
            RedrawScreen();
            break;
        case XK_c:
        case XK_period:
            Continuous = !Continuous;
            break;
        default:
            break;
        }
        break;
    case MapNotify:
    case ReparentNotify:
        break;
    default:
        if (Verbose)
            printf("Unknown event: %d\n", event->type);
    }

    /* Returning false means there was no event waiting.
     * Iterate once, if appropriate, sleep, and return.
     */
    if (Continuous) {
        Iterate(1);
        RedrawScreen();
        usleep(TickTime);
    }

    return 0;
}

